package com.tzld.piaoquan.recommend.server.service;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.base.Strings;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.server.common.enums.AppTypeEnum;
import com.tzld.piaoquan.recommend.server.gen.common.Result;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendResponse;
import com.tzld.piaoquan.recommend.server.gen.recommend.VideoProto;
import com.tzld.piaoquan.recommend.server.model.RecommendParam;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConfigService;
import com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConstants;
import com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolService;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.rank.RankResult;
import com.tzld.piaoquan.recommend.server.service.rank.RankService;
import com.tzld.piaoquan.recommend.server.service.recall.RecallParam;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import com.tzld.piaoquan.recommend.server.service.recall.RecallService;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.Region24HRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionHRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionRelative24HDupRecallStrategy;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.RegionRelative24HRecallStrategy;
import com.tzld.piaoquan.recommend.server.util.DateUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.RandomUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ZSetOperations;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.PostConstruct;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RecommendService {

    @Autowired
    private RedisTemplate<String, String> redisTemplate;


    private Map<Integer, String> ab_initial_config_map = new HashMap<>();

    @ApolloJsonValue("${ab_exp_code:{}}")
    private Map<String, Map<String, String>> abExpCodeMap = new HashMap<>();

    @ApolloJsonValue("${app_ab_code:{}}")
    private Map<String, Map<String, Map<String, String>>> appAbCodeMap = new HashMap<>();

    private int[] flowPoolIdList = {7, 8, 9, 10, 11, 12, 13, 14, 15, 16};


    @Autowired
    private FlowPoolConfigService flowPoolConfigService;

    @Autowired
    private RecallService recallService;
    @Autowired
    private RankService rankService;
    @Autowired
    private PreViewedService preViewedService;
    @Autowired
    private FlowPoolService flowPoolService;

    @PostConstruct
    public void init() {
        ab_initial_config_map.put(AppTypeEnum.VLOG.getCode(), "095");
        ab_initial_config_map.put(AppTypeEnum.LOVE_LIVE.getCode(), "144");
        ab_initial_config_map.put(AppTypeEnum.LONG_VIDEO.getCode(), "121");
        ab_initial_config_map.put(AppTypeEnum.LAO_HAO_KAN_VIDEO.getCode(), "074");
        ab_initial_config_map.put(AppTypeEnum.ZUI_JING_QI.getCode(), "069");
    }

    public RecommendResponse homepageRecommend(RecommendRequest request) {
        return recommend(request, 0);
    }

    public RecommendResponse relevantRecommend(RecommendRequest request) {
        return recommend(request, 1);
    }

    public RecommendResponse recommend(RecommendRequest request, int recommendType) {
        if (request.getVersionAuditStatus() == 1) {
            return specialMidRecommend(request);
        }
        if (StringUtils.isNotBlank(request.getMid())
                && redisTemplate.opsForSet().isMember("special:mid", request.getMid())) {
            return specialMidRecommend(request);
        }
        RecommendParam param = genRecommendParam(request, 0);
        log.info("genRecommendParam : {}", JSONUtils.toJson(param));

        List<Video> videos = videoRecommend(param);
        log.info("videoRecommend : {}", JSONUtils.toJson(videos));

        updateCache(request, param, videos);

        // 更新position
        List<VideoProto> vps = new ArrayList<>();
        for (int i = 0; i < videos.size(); i++) {
            vps.add(VideoProto.newBuilder()
                    .setPosition(i + 1)
                    .setPushFrom(Strings.nullToEmpty(videos.get(i).getPushFrom()))
                    .setAbCode(Strings.nullToEmpty(videos.get(i).getAbCode()))
                    .setVideoId(videos.get(i).getVideoId())
                    .setRovScore(videos.get(i).getRovScore())
                    .setSortScore(videos.get(i).getRovScore())
                    .setFlowPool(Strings.nullToEmpty(videos.get(i).getFlowPool()))
                    .setIsInFlowPool(videos.get(i).isInFlowPool() ? 1 : 0)
                    .setRand(videos.get(i).getRand())
                    .build());
        }

        return RecommendResponse.newBuilder()
                .setResult(Result.newBuilder().setCode(1).setMessage("success"))
                .addAllVideo(vps)
                .build();

    }

    private RecommendResponse specialMidRecommend(RecommendRequest request) {
        String keyNamePrefix = "special:videos:item:";
        String dateStr = DateUtils.getCurrentDateStr("yyyyMMdd");
        String specialKeyName = keyNamePrefix + dateStr;
        if (!redisTemplate.hasKey(specialKeyName)) {
            dateStr = DateUtils.getBeforeDaysDateStr("yyyyMMdd", 1);
            specialKeyName = keyNamePrefix + dateStr;
        }

        String lastSpecialRecallKey = String.format("recall:last:special:%s:%s:%s", request.getAppType(), request.getMid(), dateStr);
        String value = redisTemplate.opsForValue().get(lastSpecialRecallKey);
        Long idx = 0L;
        if (StringUtils.isNotBlank(value)) {
            idx = redisTemplate.opsForZSet().reverseRank(specialKeyName, value);
            idx = idx == null ? 0L : idx;
        }

        int getSize = request.getSize() * 5;
        int freq = 0;
        List<VideoProto> results = new ArrayList<>();
        while (results.size() < request.getSize()) {
            freq += 1;
            if (freq > 2) {
                break;
            }
            Set<ZSetOperations.TypedTuple<String>> data = redisTemplate.opsForZSet().reverseRangeWithScores(specialKeyName, idx, idx + getSize - 1);
            if (CollectionUtils.isEmpty(data)) {
                break;
            }
            idx += getSize;
            data.stream().forEach(t ->
                    results.add(VideoProto.newBuilder()
                            .setVideoId(Long.getLong(t.getValue(), 0L))
                            .setRovScore(t.getScore())
                            .setAbCode("99999")
                            .setPushFrom("special_mid_videos")
                            .build())
            );
        }

        if (StringUtils.isNotBlank(request.getMid()) && !CollectionUtils.isEmpty(results)) {
            redisTemplate.opsForValue().set(lastSpecialRecallKey, String.valueOf(results.get(results.size() - 1).getVideoId()), 1, TimeUnit.DAYS);
        }

        return RecommendResponse.newBuilder()
                .setResult(Result.newBuilder().setCode(1).setMessage("success"))
                .addAllVideo(results)
                .build();
    }

    private RecommendParam genRecommendParam(RecommendRequest request, int recommendType) {
        RecommendParam param = new RecommendParam();
        param.setTopK(3);
        param.setFlowPoolP(0.3);
        param.setMid(request.getMid());
        param.setUid(request.getUid());
        param.setSize(request.getSize());
        param.setVideoId(request.getVideoId());

        // client_info: {"cityCode": "", "provinceCode": "330000"}
        if (StringUtils.isNotBlank(request.getClientInfo())) {
            Map<String, String> clientInfoMap = JSONUtils.fromJson(request.getClientInfo(),
                    new TypeToken<Map<String, String>>() {
                    }, Collections.emptyMap());
            param.setCityCode(clientInfoMap.get("cityCode"));
            param.setProvinceCode(clientInfoMap.get("provinceCode"));
        }

        String abInitialConfig = ab_initial_config_map.getOrDefault(request.getAppType(), "095-1");
        Map<String, String> abExpCode = abExpCodeMap.get(abInitialConfig);
        param.setAbCode(abExpCode.get("ab_code"));
        param.setRuleKey(abExpCode.get("rule_key"));
        param.setDataKey(abExpCode.get("data_key"));
        param.setRankKeyPrefix("rank:score1:");

        // 实验配置
        if (StringUtils.isNotBlank(request.getAbExpInfo())) {
            Set<String> abExpCodes = new HashSet<>();
            Map<String, List<Map<String, String>>> map = JSONUtils.fromJson(request.getAbExpInfo(),
                    new TypeToken<Map<String, List<Map<String, String>>>>() {
                    },
                    Collections.emptyMap());
            for (List<Map<String, String>> list : map.values()) {
                list.stream().forEach(m -> abExpCodes.add(m.get("abExpCode")));
            }

            // 流量池视频分发概率
            if (abExpCodes.contains("211")) {
                param.setFlowPoolP(0.9);
            } else if (abExpCodes.contains("221")) {
                param.setFlowPoolP(0.7);
            } else if (abExpCodes.contains("299")) {
                param.setFlowPoolP(0.5);
            } else if (abExpCodes.contains("300")) {
                param.setFlowPoolP(0.4);
            } else if (abExpCodes.contains("301")) {
                param.setFlowPoolP(0.6);
            } else if (abExpCodes.contains("339")) {
                param.setFlowPoolP(0);
            }

            for (Map.Entry<String, Map<String, String>> entry : abExpCodeMap.entrySet()) {
                if (abExpCodes.contains(entry.getKey())) {
                    param.setAbCode(entry.getValue().get("ab_code"));
                    param.setRuleKey(entry.getValue().get("rule_key"));
                    param.setDataKey(entry.getValue().get("data_key"));
                    param.setRankKeyPrefix(StringUtils.isNotBlank(entry.getValue().get("rank_key_prefix"))
                            ? entry.getValue().get("rank_key_prefix")
                            : "rank:score1:");
                    break;
                }
            }
        }

        // app实验组
        // {"10018":{"eventId":"100182","ifAddedRootEventId":true},"10020":{"eventId":"100201","ifAddedRootEventId":true}}
        if (StringUtils.isNotBlank(request.getAbInfoData())) {
            Map<String, String> abInfoAppMap = new HashMap<>();
            Map<String, Map<String, String>> map = JSONUtils.fromJson(request.getAbInfoData(),
                    new TypeToken<Map<String, Map<String, String>>>() {
                    },
                    Collections.emptyMap());
            for (Map.Entry<String, Map<String, String>> entry : map.entrySet()) {
                if (MapUtils.isEmpty(entry.getValue())) {
                    continue;
                }
                String eventId = entry.getValue().get("eventId");
                if (StringUtils.isNotBlank(eventId)) {
                    abInfoAppMap.put(entry.getKey(), eventId);
                }
            }
            if (recommendType == 0) {
                String eventId = abInfoAppMap.get("10003");
                Map<String, Map<String, String>> appAbCode = appAbCodeMap.get("10003");
                if (StringUtils.isNotBlank(eventId) && appAbCode.containsKey(eventId)) {
                    param.setAbCode(appAbCode.get(eventId).get("ab_code"));
                    param.setRuleKey(appAbCode.get(eventId).get("rule_key"));
                    param.setDataKey(appAbCode.get(eventId).get("data_key"));
                }
            }

            // 流量池分发实验组划分
            int flowPoolIdChoice = flowPoolIdList[RandomUtils.nextInt(0, flowPoolIdList.length - 1)];
            param.setFlowPoolAbtestGroup("control_group");
            try {
                Map<String, List<Integer>> flowPoolConfig = flowPoolConfigService.getFlowPoolConfig();
                for (Map.Entry<String, List<Integer>> entry : flowPoolConfig.entrySet()) {
                    if (entry.getValue().contains(flowPoolIdChoice)) {
                        param.setFlowPoolAbtestGroup(entry.getKey());
                    }
                }
            } catch (Exception e) {
                log.error("flowPoolConfigCache get error", e);
            }
        }

        return param;
    }

    private List<Video> videoRecommend(RecommendParam param) {
        RecallResult recallResult = recallService.recall(convertToRecallParam(param));
        log.info("recallResult {}", recallResult);

        RankResult rankResult = rankService.rank(convertToRankParam(param, recallResult));
        log.info("rankResult {}", rankResult);

        if (rankResult == null || CollectionUtils.isEmpty(rankResult.getVideos())) {
            return Collections.emptyList();
        }

        // 只返回size条数据
        List<Video> videos = rankResult.getVideos();
        if (param.getSize() < rankResult.getVideos().size()) {
            videos = rankResult.getVideos().subList(0, param.getSize());
        }


        // mark video in flow pool
        // TODO 业务上有什么用处
        // 需要被标记的视频满足什么条件
        markFlowPoolVideo(param, videos);
        log.info("markFlowPoolVideo videos {}", videos);

        return videos;
    }

    private RecallParam convertToRecallParam(RecommendParam param) {
        RecallParam recallParam = new RecallParam();
        recallParam.setAppType(param.getAppType());
        recallParam.setAbCode(param.getAbCode());
        recallParam.setVideoId(param.getVideoId());
        recallParam.setFlowPoolAbtestGroup(param.getFlowPoolAbtestGroup());
        recallParam.setProvinceCode(param.getProvinceCode());
        recallParam.setCityCode(param.getCityCode());
        recallParam.setDataKey(param.getDataKey());
        recallParam.setRuleKey(param.getRuleKey());
        recallParam.setMid(param.getMid());
        recallParam.setSize(param.getSize());
        recallParam.setUid(param.getUid());
        return recallParam;
    }


    private RankParam convertToRankParam(RecommendParam param, RecallResult recallResult) {
        RankParam rankParam = new RankParam();
        rankParam.setRecallResult(recallResult);
        rankParam.setAbCode(param.getAbCode());
        rankParam.setSize(param.getSize());
        rankParam.setFlowPoolP(param.getFlowPoolP());
        rankParam.setTopK(param.getTopK());
        rankParam.setRankKeyPrefix(param.getRankKeyPrefix());
        rankParam.setAppType(param.getAppType());
        return rankParam;
    }

    private void markFlowPoolVideo(RecommendParam param, List<Video> videos) {
        for (Video data : videos) {
            try {
                Long videoId = data.getVideoId();
                String quick_flow_pool_isin_flow_pool_key =
                        String.format("flow:pool:quick:video:ids:%s:3", param.getAppType());
                String quick_flow_pool_flow_pool_key =
                        String.format("flow:pool:quick:video:%s:3:%s", param.getAppType(), videoId);
                if (redisTemplate.opsForSet().isMember(quick_flow_pool_isin_flow_pool_key, String.valueOf(videoId))) {
                    data.setFlowPool(redisTemplate.opsForSet().randomMember(quick_flow_pool_flow_pool_key));
                    data.setInFlowPool(true);
                } else {
                    String isIn_flow_pool_key =
                            String.format("flow:pool:video:ids:%s:3", param.getAppType());
                    String flow_pool_key =
                            String.format("flow:pool:video:%s:%s", param.getAppType(), videoId);
                    if (redisTemplate.opsForSet().isMember(isIn_flow_pool_key, String.valueOf(videoId))) {
                        data.setFlowPool(redisTemplate.opsForSet().randomMember(flow_pool_key));
                        data.setInFlowPool(true);
                    }
                }
            } catch (Exception e) {
                log.error("markFlowPoolVideo error data={}", JSONUtils.toJson(data));
            }
        }
    }

    private void updateCache(RecommendRequest request, RecommendParam param, List<Video> videos) {
        if (StringUtils.isNotBlank(request.getMid())
                || CollectionUtils.isEmpty(videos)) {
            return;
        }

        preViewedService.updateCache(request.getAppType(), request.getMid(), videos);
        updateLastVideoCache(request, param, videos);
        updateFlowPoolCache(request, param, videos);

    }

    private void updateFlowPoolCache(RecommendRequest request, RecommendParam param,
                                     List<Video> videos) {
        if (request.getAppType() == 18
                || request.getAppType() == 19
                || CollectionUtils.isEmpty(videos)) {
            return;
        }

        List<Video> flowPoolVideos = videos.stream()
                .filter(v -> v.getPushFrom().equals(FlowPoolConstants.PUSH_FORM))
                .collect(Collectors.toList());


        switch (param.getFlowPoolAbtestGroup()) {
            case FlowPoolConstants.EXPERIMENTAL_FLOW_SET_LEVEL:
                flowPoolService.updateDistributeCountWithLevel(flowPoolVideos);
                break;
            case FlowPoolConstants.EXPERIMENTAL_FLOW_SET_LEVEL_SCORE:
                flowPoolService.updateDistributeCountWithLevelScore(flowPoolVideos);
                break;
            default:
                flowPoolService.updateDistributeCountWithScore(flowPoolVideos);
                break;
        }
    }

    private void updateLastVideoCache(RecommendRequest request, RecommendParam param,
                                      List<Video> videos) {
        // 2 地域小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).getPushFrom().equals(RegionHRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(RegionHRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }

        // 3 地域24小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).getPushFrom().equals(Region24HRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(Region24HRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }

        // 4 地域相对24小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).getPushFrom().equals(RegionRelative24HRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(RegionRelative24HRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }

        // 5 地域相对24小时最后一个视频
        for (int i = videos.size() - 1; i >= 0; i--) {
            if (videos.get(i).getPushFrom().equals(RegionRelative24HDupRecallStrategy.PUSH_FORM)) {
                redisTemplate.opsForValue().set(String.format(RegionRelative24HDupRecallStrategy.LAST_VIDEO_KEY_FORMAT,
                                request.getAppType(), request.getMid()), String.valueOf(videos.get(i).getVideoId()),
                        24, TimeUnit.HOURS);
            }
        }
    }

}
