package com.tzld.piaoquan.recommend.server.service.recall.strategy;

import com.google.common.collect.Lists;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.filter.FilterResult;
import com.tzld.piaoquan.recommend.server.service.filter.RegionFilterService;
import com.tzld.piaoquan.recommend.server.service.recall.FilterParamFactory;
import com.tzld.piaoquan.recommend.server.service.recall.RecallParam;
import com.tzld.piaoquan.recommend.server.service.recall.RecallStrategy;
import com.tzld.piaoquan.recommend.server.util.DateUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ZSetOperations;

import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * @author dyp
 */
@Slf4j
public abstract class AbstractRegionRecallStrategy implements RecallStrategy {
    @Autowired
    protected RedisTemplate<String, String> redisTemplate;


    @Autowired
    protected RegionFilterService filterService;

    private final String recordKeyFormat = "{\"date\":\"%s\", \"h\":\"%d\"}";

    @Override
    public List<Video> recall(RecallParam param) {

        String recordKey = recordKey(param);
        String lastVideoKey = lastVideoKey(param);
        int idx = 0;
        String poolKey = "";

        // TODO 整段逻辑很奇怪，需要重新理解
        String record = redisTemplate.opsForValue().get(recordKey);
        log.info("recall recordKey:[{}], record:[{}]", recordKey, record);
        if (StringUtils.isNotBlank(record)) {
            Map<String, String> recordMap = JSONUtils.fromJson(record, new TypeToken<Map<String, String>>() {
            }, Collections.emptyMap());
            if (MapUtils.isNotEmpty(recordMap)) {
//                redisTemplate.delete(lastVideoKey);
//                poolKey = updateLastVideoRecord(recordKey, param);
//            } else {
                String record_dt = recordMap.get("date");
                String record_h = recordMap.get("h");
                String now_dt = DateUtils.getCurrentDateStr("yyyyMMdd");
                int h = DateUtils.getCurrentHour();
                if (record_dt.equals(now_dt) && Integer.parseInt(record_h) == h) {
                    poolKey = poolKey(param, now_dt, h);
                    idx = getIdx(lastVideoKey, poolKey);
                } else if ((record_dt.equals(now_dt) && h - Integer.parseInt(record_h) == 1)
                        || (h == 0 && Integer.parseInt(record_h) == 23)) {
                    poolKey = poolKey(param, now_dt, h);
                    if (redisTemplate.hasKey(poolKey)) {
                        redisTemplate.opsForValue().set(recordKey, String.format(recordKeyFormat, now_dt,
                                h), 2, TimeUnit.HOURS);
                        redisTemplate.delete(lastVideoKey);
                    } else {
                        poolKey = poolKey(param, record_dt, Integer.parseInt(record_h));
                        idx = getIdx(lastVideoKey, poolKey);
                    }
                } else {
                    redisTemplate.delete(lastVideoKey);
                    poolKey = updateLastVideoRecord(recordKey, param);
                }
            }
        } else {
            redisTemplate.delete(lastVideoKey);
            poolKey = updateLastVideoRecord(recordKey, param);
        }
        log.info("recall poolKey={}", poolKey);
        if (StringUtils.isBlank(poolKey)) {
            return null;
        }

        // 召回
        int getSize = param.getSize() * 5;
        int freq = 0;
        String lastVideoId = "";
        List<Video> results = new ArrayList<>();
        while (results.size() < param.getSize()) {
            freq += 1;
            if (freq > 2) {
                break;
            }
            Set<ZSetOperations.TypedTuple<String>> data = redisTemplate.opsForZSet().reverseRangeWithScores(poolKey, idx,
                    idx + getSize - 1);
            if (CollectionUtils.isEmpty(data)) {
                break;
            }
            idx += getSize;

            Map<Long, Double> videoMap = new HashMap<>();
            for (ZSetOperations.TypedTuple<String> t : data) {
                lastVideoId = t.getValue();
                videoMap.put(NumberUtils.toLong(t.getValue(), 0), t.getScore());
            }

            FilterResult filterResult = filterService.filter(FilterParamFactory.create(param, Lists.newArrayList(videoMap.keySet())));
            log.info("recall filterResult={}", JSONUtils.toJson(filterResult));

            if (filterResult != null && CollectionUtils.isNotEmpty(filterResult.getVideoIds())) {
                filterResult.getVideoIds().stream().forEach(vid -> {
                    Video recallData = new Video();
                    recallData.setVideoId(vid);
                    recallData.setAbCode(param.getAbCode());
                    recallData.setRovScore(videoMap.get(vid));
                    recallData.setPushFrom(pushFrom());
                    results.add(recallData);
                });
            }
        }


        Collections.sort(results, Comparator.comparingDouble(o -> -o.getRovScore()));

        if (StringUtils.isNotBlank(lastVideoId)
                && CollectionUtils.isNotEmpty(results)
                && StringUtils.isNotBlank(param.getMid())) {
            // 召回数据不为空 & 过滤后结果为空 &  mid不为空时，更新召回获取的末位视频id记录到定位的key中
            redisTemplate.opsForValue().set(lastVideoKey, lastVideoId, 2, TimeUnit.HOURS);
        }

        return results.subList(0, results.size() < param.getSize() ? results.size() : param.getSize());
    }

    private int getIdx(String lastVideoKey, String poolKey) {
        int idx = 0;
        String lastVideo = redisTemplate.opsForValue().get(lastVideoKey);
        if (StringUtils.isNotBlank(lastVideo)) {
            Long idxLong = redisTemplate.opsForZSet().reverseRank(poolKey, lastVideo);
            if (idxLong != null) {
                idx = idxLong.intValue();
                idx += 1;
            }
        }
        return idx;
    }

    // TODO 和 推荐排序后的lastVideo更新有什么区别?
    private String updateLastVideoRecord(String recordKey, RecallParam param) {
        String currentDateStr = DateUtils.getCurrentDateStr("yyyyMMdd");
        int h = DateUtils.getCurrentHour();

        String poolKey = poolKey(param, currentDateStr, h);
        if (!redisTemplate.hasKey(poolKey)) {
            if (h == 0) {
                currentDateStr = DateUtils.getBeforeDaysDateStr("yyyyMMdd", 1);
                h = 23;
            } else {
                h = h - 1;
            }
            poolKey = poolKey(param, currentDateStr, h);
        }
        redisTemplate.opsForValue().set(recordKey, String.format(recordKeyFormat, currentDateStr, h), 2, TimeUnit.HOURS);
        return poolKey;
    }

    abstract String recordKey(RecallParam param);

    abstract String poolKey(RecallParam param, String now_dt, int h);

    abstract String lastVideoKey(RecallParam param);
}
